@@ -9,7 +9,7 @@ from math import pi as PI
 import os
 from pathlib import Path
 import logging
-from typing import Callable, Union
+from typing import Callable, Union, Optional
 from typing import Tuple
 
 from l2hmc.configs import DynamicsConfig
@@ -225,8 +225,8 @@ class Dynamics(nn.Module):
     ) -> tuple[Tensor, dict]:
         data = self.generate_proposal_hmc(inputs, eps=eps)
         ma_, mr_ = self._get_accept_masks(data['metrics']['acc'])
-        ma = ma_.unsqueeze(-1)
-        mr = mr_.unsqueeze(-1)
+        ma = ma_.unsqueeze(-1).to(inputs[0].device)
+        mr = mr_.unsqueeze(-1).to(inputs[0].device)
 
         vout = ma * data['proposed'].v + mr * data['init'].v
         xout = ma * data['proposed'].x + mr * data['init'].x
@@ -248,6 +248,8 @@ class Dynamics(nn.Module):
         data = self.generate_proposal_fb(inputs)
 
         ma_, mr_ = self._get_accept_masks(data['metrics']['acc'])
+        ma_ = ma_.to(inputs[0].device)
+        mr_ = mr_.to(inputs[0].device)
         ma = ma_.unsqueeze(-1)
         mr = mr_.unsqueeze(-1)
 
@@ -278,8 +280,10 @@ class Dynamics(nn.Module):
         bwd = self.generate_proposal(inputs, forward=False)
 
         mf_, mb_ = self._get_direction_masks(batch_size=x.shape[0])
-        mf = mf_.unsqueeze(-1)  # mf = mf_[:, None]
-        mb = mb_.unsqueeze(-1)  # mb = mb_[:, None]
+        mf_ = mf_.to(x.device)
+        mb_ = mb_.to(x.device)
+        mf = mf_.unsqueeze(-1)  # .to(x.device)  # mf = mf_[:, None]
+        mb = mb_.unsqueeze(-1)  # .to(x.device)  # mb = mb_[:, None]
 
         v_init = mf * fwd['init'].v + mb * bwd['init'].v
 
@@ -315,14 +319,24 @@ class Dynamics(nn.Module):
         mc_states = MonteCarloStates(init=state_init,
                                      proposed=state_prop,
                                      out=state_out)
-        metrics = {
+        metrics = {}
+        for (key, vf), (_, vb) in zip(mfwd.items(), mbwd.items()):
+            try:
+                vprop = ma_ * (mf_ * vf + mb_ * vb)
+            except RuntimeError:
+                vprop = ma * (mf * vf + mb * vb)
+
+            metrics[key] = vprop
+
+        metrics.update({
             'acc': acc,
             'acc_mask': ma_,
             'sumlogdet': logdet,
             'mc_states': mc_states,
-        }
-        metrics.update(**{f'fwd/{key}': val for key, val in mfwd.items()})
-        metrics.update(**{f'bwd/{key}': val for key, val in mbwd.items()})
+        })
+
+        # metrics.update(**{f'fwd/{key}': val for key, val in mfwd.items()})
+        # metrics.update(**{f'bwd/{key}': val for key, val in mbwd.items()})
 
         return x_out, metrics
 
@@ -378,7 +392,7 @@ class Dynamics(nn.Module):
             self,
             state: State,
             logdet: Tensor,
-            step: int = None
+            step: Optional[int] = None
     ) -> dict:
         energy = self.hamiltonian(state)
         logprob = energy - logdet
@@ -522,7 +536,7 @@ class Dynamics(nn.Module):
         dh = h_init - h_prop + sumlogdet
         prob = torch.exp(
             torch.minimum(dh, torch.zeros_like(dh, device=dh.device))
-        )
+        ).to(state_init.x.device)
 
         return prob
 
@@ -530,12 +544,12 @@ class Dynamics(nn.Module):
     def _get_accept_masks(px: Tensor) -> tuple[Tensor, Tensor]:
         acc = (px > torch.rand_like(px).to(px.device)).to(torch.float)
         rej = torch.ones_like(acc) - acc
-        return acc, rej
+        return acc.to(px.device), rej.to(px.device)
 
     @staticmethod
     def _get_direction_masks(batch_size: int) -> tuple[Tensor, Tensor]:
         """Returns (forward_mask, backward_mask)."""
-        fwd = (torch.rand(batch_size).to(px.device) > 0.5).to(torch.float)
+        fwd = (torch.rand(batch_size) > 0.5).to(torch.float)
         bwd = torch.ones_like(fwd) - fwd
 
         return fwd, bwd
@@ -764,13 +778,14 @@ class Dynamics(nn.Module):
             self,
             x: Tensor,
             beta: Tensor,
-            create_graph: bool = True,
+            # create_graph: bool = True,
     ) -> Tensor:
         """Compute the gradient of the potential function."""
         x.requires_grad_(True)
         s = self.potential_energy(x, beta)
         id = torch.ones(x.shape[0], device=x.device)
         dsdx, = torch.autograd.grad(s, x,
-                                    create_graph=create_graph,
+                                    # create_graph=create_graph,
+                                    # retain_graph=True,
                                     grad_outputs=id)
         return dsdx
